# -*- coding: utf-8 -*-
# This file is part of the diagram-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

import string
from datetime import timedelta
from trytond.model import ModelSQL, ModelView, fields
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.exceptions import UserError
from trytond.i18n import gettext
from .templates import (
    template_view_point, NUM_OF_GRAPH, sel_scaling_chart, sel_scaling_info)
from .colors import sel_color as sel_bgcolor


sel_numdays = [
    ('30', '30 Days'),
    ('60', '2 Months'),
    ('90', '3 Months'),
    ('180', '6 Months'),
    ('365', '1 Year'),
    ]

sel_maincolor = [
    ('default', 'Default'),
    ('red', 'Red'),
    ('green', 'Green'),
    ('grey', 'Grey'),
    ('black', 'Black'),
    ('darkcyan', 'Dark Cyan'),
]


class Chart(ModelSQL, ModelView):
    'Chart'
    __name__ = 'diagram.chart'

    company = fields.Many2One(
        string='Company', model_name='company.company',
        required=True, ondelete="RESTRICT")
    name = fields.Char(string='Name', required=True)
    graphdef = fields.One2Many(
        string='Graphs', required=True,
        help='Definition of the displayed data', field='chart',
        model_name='diagram.graphdef', size=NUM_OF_GRAPH,
        order=[('sequence', 'ASC')])
    legend = fields.Boolean(
        string='Legend', help='Shows a legend on the chart.')
    bgcolor = fields.Selection(
        string='Background Color', required=True,
        help='Background color of the chart area.', sort=False,
        selection=sel_bgcolor)
    maincolor = fields.Selection(
        string='Color scheme', required=True,
        help='The color scheme determines the hue of ' +
        'all components of the chart.',
        selection=sel_maincolor, sort=False)
    lines = fields.One2Many(
        string='Chart', readonly=True,
        model_name='diagram.point', field='chart')
    numdays = fields.Selection(
        string='Interval', required=True,
        selection=sel_numdays, sort=False)
    start_date = fields.Date(string='Start Date')
    ui_view_point = fields.Many2One(
        string='UI View Point',
        model_name='ir.ui.view', ondelete='SET NULL')
    scaling = fields.Selection(
        string='Scaling', required=True,
        selection=sel_scaling_chart, sort=True)
    scaling_info = fields.Function(fields.Selection(
        string='Scaling',
        selection=sel_scaling_info), 'on_change_with_scaling_info')

    @classmethod
    def default_scaling(cls):
        """ default: off
        """
        return 'off'

    @classmethod
    def default_maincolor(cls):
        """ default: 'default'
        """
        return 'default'

    @classmethod
    def default_bgcolor(cls):
        """ default: Yellow 5
        """
        return '#ffffc0'

    @classmethod
    def default_legend(cls):
        """ default: True
        """
        return True

    @staticmethod
    def default_company():
        return Transaction().context.get('company') or None

    @staticmethod
    def default_numdays():
        return '90'

    def limit_start_date(self, start_date):
        """ update start-date if end-date wo
        """
        IrDate = Pool().get('ir.date')

        if start_date is None:
            return None

        td1 = timedelta(days=int(self.numdays))
        if (start_date + td1) > IrDate.today():
            start_date = IrDate.today() - td1
        return start_date

    def used_start_date(self):
        """ get start-date by context, chart, today
        """
        IrDate = Pool().get('ir.date')
        context = Transaction().context

        start_date = context.get('point_date', None)
        if start_date is None:
            start_date = self.start_date
            if start_date is None:
                start_date = IrDate.today() - \
                    timedelta(days=int(self.numdays))
        result = self.limit_start_date(start_date)
        return result

    def used_end_date(self):
        """ end-date by start_date + numdays
        """
        return self.used_start_date() + timedelta(days=int(self.numdays))

    @fields.depends('scaling')
    def on_change_with_scaling_info(self, name=None):
        """ get scaling setting
        """
        if self.scaling:
            return self.scaling

    @classmethod
    def uiview_create(cls, charts):
        """ create view for current setup of chart
        """
        pool = Pool()
        UiView = pool.get('ir.ui.view')
        Chart2 = pool.get('diagram.chart')

        cls.uiview_delete(charts)
        to_write_chart = []
        for chart in charts:
            if len(chart.graphdef) == 0:
                continue

            view_point, = UiView.create([{
                'model': 'diagram.point',
                'module': 'diagram',
                'priority': 10,
                'type': 'graph',
                'data': template_view_point % {
                    'bgcol': '' if chart.bgcolor == 'default'
                        else 'background="%s"' % chart.bgcolor,
                    'colscheme': '' if chart.maincolor == 'default'
                        else 'color="%s"' % chart.maincolor,
                    'legend': '1' if chart.legend is True else '0',
                    'lines': '\n'.join([
                        x.line_html() for x in chart.graphdef]),
                    }}])

            to_write_chart.extend([
                [chart], {'ui_view_point': view_point.id}])

        if len(to_write_chart) > 0:
            Chart2.write(*to_write_chart)

    @classmethod
    def uiview_delete(cls, charts):
        """ delete action view from chart
        """
        pool = Pool()
        UiView = pool.get('ir.ui.view')

        to_delete_uiview = []
        for chart in charts:
            if chart.ui_view_point:
                to_delete_uiview.append(chart.ui_view_point)

        if len(to_delete_uiview) > 0:
            UiView.delete(to_delete_uiview)

    @classmethod
    def validate(cls, charts):
        """ check: bg-color
        """
        super(Chart, cls).validate(charts)

        for chart in charts:
            if (len(chart.bgcolor) != 7) or \
                    (not chart.bgcolor.startswith('#')):
                raise UserError(gettext(
                    'diagram.msg_invalid_bgcolor',
                    chname=chart.rec_name))
            if not all(c in string.hexdigits for c in chart.bgcolor[1:]):
                raise UserError(gettext(
                    'diagram.msg_invalid_bgcolor',
                    chname=chart.rec_name))

    @classmethod
    def create(cls, vlist):
        """ add points
        """
        GraphDef = Pool().get('diagram.graphdef')

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if 'numdays' in values.keys():
                values['lines'] = [('create', [{
                        'sequence': x,
                    } for x in range(int(values['numdays']))])]

        with Transaction().set_context({
                'chart_skip_viewupdate': True}):
            records = super(Chart, cls).create(vlist)

        cls.uiview_create(records)
        GraphDef.update_scaling_from_parent([
            y for x in records for y in x.graphdef])
        return records

    @classmethod
    def write(cls, *args):
        """ update points
        """
        GraphDef = Pool().get('diagram.graphdef')

        actions = iter(args)
        to_write = []
        to_update_view = []
        to_update_graph = []
        for lines, values in zip(actions, actions):
            if 'numdays' in values.keys():
                for line in lines:
                    if line.numdays != values['numdays']:
                        values['lines'] = [
                            ('delete', [x.id for x in line.lines]),
                            ('create', [{
                                'sequence': x,
                                } for x in range(int(values['numdays']))]),
                            ]
                    to_write.extend([[line], values])
            elif len(set({
                'graphdef', 'name', 'maincolor',
                    'bgcolor', 'legend'}).intersection(values.keys())) > 0:
                to_update_view.extend(lines)
                to_write.extend([lines, values])
            else:
                to_write.extend([lines, values])

            if 'scaling' in values.keys():
                to_update_graph.extend(lines)

        with Transaction().set_context({
                'chart_skip_viewupdate': True}):
            super(Chart, cls).write(*to_write)

        GraphDef.update_scaling_from_parent([
            y for x in to_update_graph for y in x.graphdef])

        if len(to_update_view) > 0:
            cls.uiview_create(to_update_view)
        GraphDef._scaling_graph_cache.clear_all()

    @classmethod
    def delete(cls, charts):
        """ delete views
        """
        cls.uiview_delete(charts)
        with Transaction().set_context({
                'chart_skip_viewupdate': True}):
            super(Chart, cls).delete(charts)

# end Chart
