# -*- coding: utf-8 -*-
# This file is part of the diagram-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

import html
from decimal import Decimal
from datetime import timedelta
from trytond.model import ModelSQL, ModelView, fields, sequence_ordered
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.pyson import Eval
from trytond.cache import Cache
from .colors import sel_color as sel_linecolor
from . templates import template_view_line, sel_scaling_graph, sel_scaling_info

SCALE_MIN = Decimal('0.0')
SCALE_MAX = Decimal('100.0')


class GraphDef(sequence_ordered(), ModelSQL, ModelView):
    'Graph Definition'
    __name__ = 'diagram.graphdef'

    chart = fields.Many2One(
        string='Chart', ondelete='CASCADE',
        model_name='diagram.chart', required=True)
    name = fields.Char(
        string='Name', required=True, help='visible Name in Chart')

    linecolor = fields.Selection(
        string='Color', required=True,
        help="Color of line, for automatic coloring use 'Default'",
        sort=False, selection=sel_linecolor)
    linefill = fields.Boolean(
        string='Fill',
        help='Fills the area below the line. Ignored by the web client.')
    scaling = fields.Selection(
        string='Scaling', required=True,
        selection=sel_scaling_graph, sort=False,
        states={
            'readonly': Eval('parent_scaling', '') != 'graph',
        }, depends=['parent_scaling'])
    scaling_info = fields.Function(fields.Selection(
        string='Scaling',
        selection=sel_scaling_info), 'on_change_with_scaling_info')
    scaling_info_string = scaling_info.translated('scaling_info')
    parent_scaling = fields.Function(fields.Char(
        string='Scaling',
        readonly=True), 'on_change_with_parent_scaling')

    dtype = fields.Selection(
        string='Type', required=True,
        help='Data type to be displayed', selection='get_dtypes')
    dtype_string = dtype.translated('dtype')
    currency = fields.Many2One(
        string='Currency', model_name='currency.currency',
        domain=[('has_rates', '=', True)],
        states={
            'invisible': Eval('dtype', '') != 'currency.currency',
            'required': Eval('dtype', '') == 'currency.currency',
        }, depends=['dtype'])

    _scaling_graph_cache = Cache(
        'diagram.chart.scaling',
        duration=timedelta(seconds=3600))

    @classmethod
    def default_scaling(cls):
        """ default: off
        """
        return 'off'

    @classmethod
    def default_linefill(cls):
        """ default: not filled
        """
        return False

    @classmethod
    def default_linecolor(cls):
        """ default: automatic-color
        """
        return 'default'

    @classmethod
    def default_currency(cls):
        """ currency of company
        """
        Company = Pool().get('company.company')

        company = Transaction().context.get('company')
        if company:
            company = Company(company)
            if company.currency:
                return company.currency.id

    @fields.depends('chart', '_parent_chart.scaling')
    def on_change_with_parent_scaling(self, name=None):
        """ get parent-scaling
        """
        if self.chart:
            return self.chart.scaling

    @fields.depends('scaling')
    def on_change_with_scaling_info(self, name=None):
        """ get scaling setting
        """
        if self.scaling:
            return self.scaling

    def line_html(self):
        """ generate <field .../>
        """
        def get_numcnt(chart, grdef):
            """ get number of graphdef
            """
            cnt1 = 1
            for x in range(len(chart.graphdef)):
                if chart.graphdef[x].id == grdef.id:
                    break
                cnt1 += 1
            return cnt1

        return template_view_line % {
            'string': html.escape(self.name),
            'fname': 'graph%d' % get_numcnt(self.chart, self),
            'fill': '0' if self.linefill is False else '1',
            'linecolor': '' if self.linecolor == 'default'
            else 'color="%s"' % self.linecolor}

    @classmethod
    def default_dtype(cls):
        """ default: first item
        """
        return cls._get_dtypes()[0]

    @classmethod
    def _get_dtypes(cls):
        """ return list of types
        """
        return ['currency.currency']

    def get_recname_value(self):
        """ value by dtype
        """
        if self.dtype == 'currency.currency':
            return getattr(self.currency, 'rec_name', '-')
        return '-'

    def get_rec_name(self, name):
        """ type, data, setting
        """
        g_setting = []
        g_setting.append(getattr(self, 'scaling_info_string', '-'))
        return '%(type)s: %(value)s [%(setting)s]' % {
            'type': getattr(self, 'dtype_string', '-'),
            'value': self.get_recname_value(),
            'setting': ','.join(g_setting)}

    @classmethod
    def get_dtypes(cls):
        """ get list for selection
        """
        Model = Pool().get('ir.model')

        get_name = Model.get_name
        models = cls._get_dtypes()
        return [(None, '')] + [(m, get_name(m)) for m in models]

    def compute_scaling_factor(self, min_val, max_val):
        """ generate scaling values from min/max
        """
        GraphDef2 = Pool().get('diagram.graphdef')

        if (min_val is None) or (max_val is None):
            return None

        if (max_val - min_val) == Decimal('0.0'):
            # move line to 50%
            scale_params = {
                'factor': Decimal('1.0'),
                'offset': (SCALE_MAX - SCALE_MIN) / Decimal('2.0') - min_val,
                'min': SCALE_MIN,
                'max': SCALE_MAX}
        else:
            factor = (SCALE_MAX - SCALE_MIN) / (max_val - min_val)
            scale_params = {
                'factor': factor.quantize(Decimal('0.0001')),
                'offset': (SCALE_MIN - min_val) * factor,
                'min': SCALE_MIN,
                'max': SCALE_MAX}
        GraphDef2._scaling_graph_cache.set(
                self.get_graph_cache_key(),
                scale_params)
        return scale_params

    def get_scaling_for_currency_currency(self):
        """ get scaling for currency
        """
        pool = Pool()
        Rate = pool.get('currency.currency.rate')

        if self.scaling == 'fix':
            return None

        if self.scaling == 'alldata':
            query = [('currency.id', '=', self.currency.id)]
        elif self.scaling == 'view':
            query = [
                ('currency.id', '=', self.currency.id),
                ('date', '>=', self.chart.used_start_date()),
                ('date', '<=', self.chart.used_end_date()),
                ]
        elif self.scaling == 'six':
            query = [
                ('currency.id', '=', self.currency.id),
                ('date', '>=',
                    self.chart.used_start_date() - timedelta(days=180)),
                ('date', '<=', self.chart.used_end_date()),
                ]

        min_rec = Rate.search(query, limit=1, order=[('rate', 'ASC')])
        max_rec = Rate.search(query, limit=1, order=[('rate', 'DESC')])
        min_val = min_rec[0].rate if len(min_rec) > 0 else None
        max_val = max_rec[0].rate if len(max_rec) > 0 else None

        return self.compute_scaling_factor(min_val, max_val)

    def get_graph_cache_key(self):
        """ key for caching settings
        """
        return '%(id)d-%(dtype)s-%(scaling)s-%(date)s-%(days)s' % {
            'id': self.id,
            'dtype': self.dtype,
            'scaling': self.scaling,
            'date': self.chart.used_start_date().isoformat(),
            'days': self.chart.numdays}

    def get_scaling_params(self):
        """ get scaling-params for current graph
        """
        GraphDef2 = Pool().get('diagram.graphdef')

        cache_key = self.get_graph_cache_key()
        scale_params = GraphDef2._scaling_graph_cache.get(cache_key)
        if scale_params is not None:
            return scale_params

        if self.scaling == 'off':
            return None

        scale_meth = getattr(
            self, 'get_scaling_for_%s' % self.dtype.replace('.', '_'),
            None)

        if scale_meth is not None:
            return scale_meth()
        else:
            raise ValueError('scaling for %s not implemented' % self.dtype)

    def get_field_key(self):
        """ get to read value from json
        """
        if self.dtype == 'currency.currency':
            return 'cur%d' % self.currency.id
        return 'none'

    @classmethod
    def update_chart_view(cls, records):
        """ rewrite views
        """
        Chart = Pool().get('diagram.chart')

        if Transaction().context.get('chart_skip_viewupdate', False) is False:
            charts = []
            for record in records:
                if record.chart not in charts:
                    charts.append(record.chart)
            Chart.uiview_create(charts)

    @classmethod
    def update_scaling_from_parent(cls, graphs):
        """ update local scaling if parent=graph
        """
        GraphDef2 = Pool().get('diagram.graphdef')

        to_write = []
        for graph in graphs:
            if (graph.chart.scaling != 'graph') and \
                    (graph.chart.scaling != graph.scaling):
                to_write.extend([
                    [graph], {'scaling': graph.chart.scaling}])
        if to_write:
            GraphDef2.write(*to_write)

    @classmethod
    def create(cls, vlist):
        """ clear keyname-cache
        """
        records = super(GraphDef, cls).create(vlist)
        cls.update_chart_view(records)
        cls.update_scaling_from_parent(records)
        return records

    @classmethod
    def write(cls, *args):
        """ clear keyname-cache
        """
        super(GraphDef, cls).write(*args)

        actions = iter(args)
        records = []
        for lines, values in zip(actions, actions):
            records.extend(lines)
        cls.update_chart_view(records)
        cls.update_scaling_from_parent(records)
        cls._scaling_graph_cache.clear_all()

    @classmethod
    def delete(cls, graphs):
        """ delete views
        """
        super(GraphDef, cls).delete(graphs)
        cls.update_chart_view(graphs)

# end GraphDef
