# -*- coding: utf-8 -*-
# This file is part of the diagram-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

from trytond.model import ModelSQL, ModelView, fields, sequence_ordered
from trytond.pool import Pool
from trytond.transaction import Transaction
from decimal import Decimal
from sql.functions import CurrentDate
from sql.conditionals import Coalesce
from sql.functions import Function
from .templates import NUM_OF_GRAPH


class JsonObjAgg(Function):
    """ build json object from columns
    """
    __slots__ = ()
    _function = 'json_object_agg'

# end JsonObjAgg


class Concat2(Function):
    """ concat columns
    """
    __slots__ = ()
    _function = 'concat'

# end Concat2


class ChartPoint(sequence_ordered(), ModelSQL, ModelView):
    'Chart Point'
    __name__ = 'diagram.point'

    chart = fields.Many2One(
        string='Chart', required=True,
        ondelete='CASCADE', model_name='diagram.chart')
    date = fields.Function(fields.Date(
        string='Date', readonly=True),
        'get_graph_data', searcher='search_date')

    @classmethod
    def __setup__(cls):
        super(ChartPoint, cls).__setup__()
        cls.add_graph_fields()

    @classmethod
    def fields_view_get(cls, view_id, view_type='form'):
        """ replace form-view-id
        """
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        Chart = pool.get('diagram.chart')
        context = Transaction().context

        # get id of origin chart-form
        form_id = ModelData.get_id('diagram', 'point_view_graph')

        # active_chart was added by tree_open-action
        active_chart = context.get('active_chart', None)

        # check if we are requested for our default form...
        if (view_type == 'graph') and (view_id == form_id) and \
                (active_chart is not None):
            chart, = Chart.browse([active_chart])
            if chart.ui_view_point:
                # ... switch to view, created by chart-config
                view_id = chart.ui_view_point.id

        return super(ChartPoint, cls).fields_view_get(
                    view_id=view_id, view_type=view_type)

    @classmethod
    def add_graph_fields(cls):
        """ add fields to Point-class
        """
        for f_nr in range(1, NUM_OF_GRAPH + 1):
            setattr(
                cls, 'graph%d' % f_nr,
                fields.Function(fields.Numeric(
                    string='Graph %d' % f_nr,
                    readonly=True,
                    digits=(10, 4)),
                    'get_graph_data'))

    @staticmethod
    def order_date(tables):
        """ order by date
        """
        Point2 = Pool().get('diagram.point')

        (tab_date, tab_point) = Point2.get_sqltable_dates()
        table, _ = tables[None]

        query = tab_date.select(
                tab_date.date,
                where=tab_date.id == table.id)
        return [query]

    @classmethod
    def search_date(cls, names, clause):
        """ search in 'date'
        """
        Operator = fields.SQL_OPERATORS[clause[1]]
        (tab_date, tab_point) = cls.get_sqltable_dates()

        query = tab_date.select(
                tab_date.id,
                where=Operator(tab_date.date, clause[2]))
        return [('id', 'in', query)]

    @classmethod
    def get_table_parts(cls):
        """ return a list of tables to union,
            table must contain the columns:
                date, key, val
        """
        pool = Pool()
        Rate = pool.get('currency.currency.rate')
        tab_rate = Rate.__table__()

        # list of currency rates/dates/ids
        tab_allcur = tab_rate.select(
                tab_rate.date,
                Concat2('cur', tab_rate.currency).as_('key'),
                tab_rate.rate.as_('val'))
        return [tab_allcur]

    @classmethod
    def get_sqltable_dates(cls):
        """ get table: id + date
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')
        Point = pool.get('diagram.point')
        tab_chart = Chart.__table__()
        tab_point = Point.__table__()
        context = Transaction().context

        # list of id and date
        # generate date from start_date and sequence
        tab_date = tab_chart.join(
                tab_point,
                condition=tab_chart.id == tab_point.chart,
            ).select(
                tab_point.id,
                (Coalesce(      # use start-date in following order:
                    context.get('point_date', None),    # 1. from context
                    tab_chart.start_date,               # 2. from chart
                    CurrentDate() -                     # 3. current-date
                    Coalesce(tab_chart.numdays, '90').cast('integer'),
                    ) + tab_point.sequence).cast('date').as_('date'),
            )
        return (tab_date, tab_point)

    @classmethod
    def scale_point_value(cls, value, scaling):
        """ scale value of point, apply limits
        """
        if value is None:
            return None

        result = (
                value * scaling.get('factor', Decimal('1.0')) +
                scaling.get('offset', Decimal('0.0'))
            ).quantize(Decimal('0.0001'))

        maxval = scaling.get('max', None)
        if maxval is not None:
            result = result if result < maxval else maxval

        minval = scaling.get('min', None)
        if minval is not None:
            result = result if result > minval else minval

        return result

    @classmethod
    def get_interpolated_val(cls, keyname, query_date):
        """ query two neighbour-values to
            interpolate missing value
        """
        Rate = Pool().get('currency.currency.rate')

        if keyname is None:
            return None

        # check if query is for us
        if keyname.startswith('cur'):
            curreny_id = int(keyname[len('cur'):])

            before = Rate.search([
                ('date', '<', query_date),
                ('currency.id', '=', curreny_id),
                ], limit=1, order=[('date', 'DESC')])

            after = Rate.search([
                ('date', '>', query_date),
                ('currency.id', '=', curreny_id),
                ], limit=1, order=[('date', 'ASC')])

            if (len(before) == 1) and (len(after) == 1):
                result = cls.interpolate_linear(
                        (after[0].date, after[0].rate),
                        (before[0].date, before[0].rate),
                        query_date)
                return result
            elif len(before) == 1:
                return before[0].rate
            elif len(after) == 1:
                return after[0].rate
        return None

    @classmethod
    def interpolate_linear(cls, p2, p1, q_date):
        """ p1, p2: (date, value)
            q_date: query-date
            y = y1 + (x-x1)(y2-y1)/(x2-x1)
        """
        (date2, val2) = p2
        (date1, val1) = p1

        d1 = 0
        d2 = (date2 - date1).days
        q_d = (q_date - date1).days

        if d1 == d2:
            result = (val1 + val2) / Decimal('2.0')
        else:
            result = val1 + (q_d - d1) * (val2 - val1) / (d2 - d1)
        return result.quantize(Decimal('0.0001'))

    @classmethod
    def get_graph_data(cls, points, names):
        """ get point-values
        """
        cursor = Transaction().connection.cursor()

        # set all points to None
        result = {name: {y.id: None for y in points} for name in names}
        records = {x.id: x for x in points}

        (tab_date, tab_point) = cls.get_sqltable_dates()
        tab_date.where = tab_point.id.in_(list(records.keys()))

        # build union from different data sources
        tables = cls.get_table_parts()
        tab_union = tables[0]
        for tab2 in tables[1:]:
            tab_union = tab_union | tab2

        # build query
        tab_ges = tab_date.join(
                tab_union,
                condition=tab_union.date == tab_date.date,
                type_='LEFT OUTER',
            ).select(
                tab_date.id,
                tab_date.date,
                JsonObjAgg(
                    Coalesce(tab_union.key, 'none'),
                    tab_union.val,
                ).as_('keyvalue'),
                group_by=[tab_date.id, tab_date.date])
        cursor.execute(*tab_ges)
        point_lines = cursor.fetchall()

        for pline in point_lines:
            (id1, date1, keyval) = pline

            for name in names:
                if name == 'date':
                    result['date'][id1] = date1
                else:
                    (key_name, scaling) = records[id1].get_keyname(name)
                    val = keyval.get(key_name, None)
                    if val is not None:
                        result[name][id1] = \
                            Decimal(val).quantize(Decimal('0.0001'))
                    else:
                        # interpolate missing values
                        result[name][id1] = \
                            cls.get_interpolated_val(key_name, date1)

                    if scaling is not None:
                        result[name][id1] = cls.scale_point_value(
                                result[name][id1],
                                scaling)
        return result

    def get_keyname(self, field_name):
        """ get key to read value from json
        """
        if self.chart:
            graph_number = int(field_name[len('graph'):]) - 1
            if not (0 <= graph_number < len(self.chart.graphdef)):
                return (None, None)

            keyname = self.chart.graphdef[graph_number].get_field_key()
            scaling = self.chart.graphdef[graph_number].get_scaling_params()

            return (keyname, scaling)
        return ('none', None)

# end ChartPoint
