# -*- coding: utf-8 -*-
# This file is part of the diagram-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

from datetime import date, timedelta
from trytond.tests.test_tryton import ModuleTestCase, with_transaction
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.exceptions import UserError
from trytond.modules.company.tests import create_company
from trytond.modules.diagram.templates import \
    template_view_point, template_view_line
from decimal import Decimal


class ChartTestCase(ModuleTestCase):
    'Test cashbook reporting module'
    module = 'diagram'

    def prep_company(self):
        """ get/create company
        """
        Company = Pool().get('company.company')

        company = Company.search([])
        if len(company) > 0:
            company = company[0]
        else:
            company = create_company(name='m-ds')
        return company

    def prep_2nd_currency(self, company):
        """ add EUR as 2nd currency
        """
        pool = Pool()
        Currency = pool.get('currency.currency')
        CurrencyRate = pool.get('currency.currency.rate')
        Company = pool.get('company.company')

        usd, = Currency.search([('name', '=', 'usd')])
        euros = Currency.search([('code', '=', 'EUR')])

        if len(euros) == 0:
            euro, = Currency.create([{
                'name': 'Euro',
                'symbol': '€',
                'code': 'EUR',
                'numeric_code': '978',
                'rounding': Decimal('0.01'),
                'digits': 2,
                }])
        else:
            euro = euros[0]
        self.assertEqual(euro.has_rates, False)
        self.assertEqual(usd.has_rates, True)
        self.assertEqual(Currency.search_count([('has_rates', '=', True)]), 1)
        self.assertEqual(Currency.search_count([('has_rates', '=', False)]), 1)

        # set company-currency to euro
        self.assertEqual(company.currency.name, 'usd')
        Company.write(*[
            [company],
            {
                'currency': euro.id,
            }])
        self.assertEqual(company.currency.name, 'Euro')

        # add rate for euro/usd @ 05/02/2022
        # EUR is base-currency
        CurrencyRate.create([{
            'date': date(2022, 5, 2),
            'currency': euro.id,
            'rate': Decimal('1.0'),
            }, {
            'date': date(2022, 5, 2),
            'currency': usd.id,
            'rate': Decimal('1.05'),
            }])

        self.assertEqual(euro.has_rates, True)
        self.assertEqual(usd.has_rates, True)
        self.assertEqual(Currency.search_count([('has_rates', '=', True)]), 2)
        self.assertEqual(Currency.search_count([('has_rates', '=', False)]), 0)

        # delete unwanted rates
        usd_1 = CurrencyRate.search([
            ('currency.id', '=', usd.id),
            ('date', '!=', date(2022, 5, 2)),
            ])
        CurrencyRate.delete(usd_1)
        return (usd, euro)

    @with_transaction()
    def test_reporting_func_interpolate_linear(self):
        """ check interpolate function
        """
        Point = Pool().get('diagram.point')
        self.assertEqual(Point.interpolate_linear(
                    (date(2022, 5, 9), Decimal('10.0')),
                    (date(2022, 5, 1), Decimal('5.0')),
                    date(2022, 5, 5)
                ),
                Decimal('7.5'))

        self.assertEqual(Point.interpolate_linear(
                    (date(2022, 5, 9), Decimal('10.0')),
                    (date(2022, 5, 1), Decimal('5.0')),
                    date(2022, 5, 1)
                ),
                Decimal('5.0'))

        self.assertEqual(Point.interpolate_linear(
                    (date(2022, 5, 9), Decimal('10.0')),
                    (date(2022, 5, 1), Decimal('5.0')),
                    date(2022, 5, 9)
                ),
                Decimal('10.0'))

        self.assertEqual(Point.interpolate_linear(
                    (date(2022, 5, 9), Decimal('10.0')),
                    (date(2022, 5, 1), Decimal('5.0')),
                    date(2022, 5, 8)
                ),
                Decimal('9.375'))

    @with_transaction()
    def test_reporting_interpolated_chart(self):
        """ create chart, add graphdef, set interval
            interpolate some values
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')
        Point = pool.get('diagram.point')
        Currency = pool.get('currency.currency')

        company = self.prep_company()
        (usd, euro) = self.prep_2nd_currency(company)
        self.assertEqual(len(usd.rates), 1)
        self.assertEqual(usd.rates[0].date, date(2022, 5, 2))
        self.assertEqual(usd.rates[0].rate, Decimal('1.05'))

        Currency.write(*[
            [usd],
            {
                'rates': [('create', [{
                    'date': date(2022, 5, 10),
                    'rate': Decimal('1.2'),
                    }])],
            }])

        with Transaction().set_context({
                'company': company.id}):
            chart, = Chart.create([{
                'name': 'Chart 1',
                'numdays': '90',
                'start_date': date(2022, 4, 1),
                'graphdef': [('create', [{
                        'name': 'Euro',
                        'dtype': 'currency.currency',
                        'currency': euro.id,
                    }, {
                        'name': 'US$',
                        'dtype': 'currency.currency',
                        'currency': usd.id,
                    }])],
                }])

            self.assertEqual(chart.name, 'Chart 1')
            self.assertEqual(chart.numdays, '90')
            self.assertEqual(chart.start_date, date(2022, 4, 1))

            points = Point.search([('date', '=', date(2022, 5, 2))])
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0].date, date(2022, 5, 2))
            self.assertEqual(points[0].graph1, Decimal('1.0'))
            self.assertEqual(points[0].graph2, Decimal('1.05'))

            points = Point.search([('date', '=', date(2022, 5, 10))])
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0].date, date(2022, 5, 10))
            self.assertEqual(points[0].graph2, Decimal('1.2'))

            points = Point.search([('date', '=', date(2022, 5, 3))])
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0].date, date(2022, 5, 3))
            self.assertEqual(points[0].graph2, Decimal('1.0688'))

            points = Point.search([('date', '=', date(2022, 5, 5))])
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0].date, date(2022, 5, 5))
            self.assertEqual(points[0].graph2, Decimal('1.1062'))

    @with_transaction()
    def test_reporting_chart_view(self):
        """ create chart, add graphdef, check created view
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')

        company = self.prep_company()
        (usd, euro) = self.prep_2nd_currency(company)
        self.assertEqual(len(usd.rates), 1)
        self.assertEqual(usd.rates[0].date, date(2022, 5, 2))
        self.assertEqual(usd.rates[0].rate, Decimal('1.05'))

        with Transaction().set_context({
                'company': company.id}):
            chart, = Chart.create([{
                'name': 'Chart 1',
                'numdays': '90',
                'start_date': date(2022, 4, 1),
                'graphdef': [('create', [{
                        'name': 'Euro',
                        'dtype': 'currency.currency',
                        'currency': euro.id,
                    }, {
                        'name': 'US$',
                        'dtype': 'currency.currency',
                        'currency': usd.id,
                        'linecolor': '#f68e76',
                        'linefill': True,
                    }])],
                }])

            self.assertEqual(chart.name, 'Chart 1')
            self.assertEqual(chart.numdays, '90')
            self.assertEqual(chart.start_date, date(2022, 4, 1))
            self.assertEqual(len(chart.graphdef), 2)
            self.assertEqual(chart.graphdef[0].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[0].currency.rec_name, 'Euro')
            self.assertEqual(chart.graphdef[1].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[1].currency.rec_name, 'usd')

            self.assertEqual(
                chart.ui_view_point.data,
                template_view_point % {
                    'bgcol': 'background="#ffffc0"',
                    'legend': '1',
                    'colscheme': '',
                    'lines': '\n'.join([
                        template_view_line % {
                            'string': 'Euro',
                            'fname': 'graph1',
                            'fill': '0',
                            'linecolor': '',
                            },
                        template_view_line % {
                            'string': 'US$',
                            'fname': 'graph2',
                            'fill': '1',
                            'linecolor': 'color="#f68e76"'},
                        ]),
                    })

    @with_transaction()
    def test_reporting_bgvalidate(self):
        """ create chart, check validation of color
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')

        company = self.prep_company()

        with Transaction().set_context({
                'company': company.id}):
            chart, = Chart.create([{
                'name': 'Chart 1',
                'numdays': '90',
                'start_date': date(2022, 4, 1),
                'graphdef': [('create', [{
                        'name': 'Euro',
                        'dtype': 'currency.currency',
                        'currency': company.currency.id,
                    }])],
                }])

            self.assertEqual(chart.name, 'Chart 1')
            self.assertEqual(chart.bgcolor, '#ffffc0')

            Chart.write(*[
                [chart],
                {
                    'bgcolor': '#e3d200',
                }])
            self.assertEqual(chart.bgcolor, '#e3d200')

            self.assertRaisesRegex(
                UserError,
                'The value "#3s5678" for field "Background Color" in ' +
                '"Chart 1" of "Chart" is not one of the allowed options.',
                Chart.write,
                *[
                    [chart],
                    {
                        'bgcolor': '#3s5678',
                    },
                ])

    @with_transaction()
    def test_reporting_chart(self):
        """ create chart, add graphdef, set interval
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')
        Point = pool.get('diagram.point')

        company = self.prep_company()
        (usd, euro) = self.prep_2nd_currency(company)
        self.assertEqual(len(usd.rates), 1)
        self.assertEqual(usd.rates[0].date, date(2022, 5, 2))
        self.assertEqual(usd.rates[0].rate, Decimal('1.05'))

        with Transaction().set_context({
                'company': company.id}):
            chart, = Chart.create([{
                'name': 'Chart 1',
                'numdays': '90',
                'start_date': date(2022, 4, 1),
                'graphdef': [('create', [{
                        'name': 'Euro',
                        'dtype': 'currency.currency',
                        'currency': euro.id,
                    }, {
                        'name': 'US$',
                        'dtype': 'currency.currency',
                        'currency': usd.id,
                    }])],
                }])

            self.assertEqual(chart.name, 'Chart 1')
            self.assertEqual(chart.numdays, '90')
            self.assertEqual(chart.start_date, date(2022, 4, 1))
            self.assertEqual(chart.bgcolor, '#ffffc0')
            self.assertEqual(chart.legend, True)
            self.assertEqual(len(chart.graphdef), 2)
            self.assertEqual(chart.graphdef[0].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[0].currency.rec_name, 'Euro')
            self.assertEqual(chart.graphdef[1].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[1].currency.rec_name, 'usd')

            self.assertEqual(len(chart.lines), 90)
            self.assertEqual(chart.lines[0].date, date(2022, 4, 1))
            # interpolated values
            self.assertEqual(chart.lines[0].graph1, Decimal('1.0'))
            self.assertEqual(chart.lines[0].graph2, Decimal('1.05'))

            self.assertEqual(chart.lines[31].date, date(2022, 5, 2))
            self.assertEqual(chart.lines[31].graph1, Decimal('1.0'))
            self.assertEqual(chart.lines[31].graph2, Decimal('1.05'))

            points = Point.search([('date', '=', date(2022, 5, 2))])
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0].date, date(2022, 5, 2))
            self.assertEqual(points[0].graph1, Decimal('1.0'))
            self.assertEqual(points[0].graph2, Decimal('1.05'))

            points = Point.search([
                    ('date', '>=', date(2022, 5, 1)),
                    ('date', '<=', date(2022, 5, 31)),
                ], order=[('date', 'ASC')])
            self.assertEqual(len(points), 31)
            self.assertEqual(points[0].date, date(2022, 5, 1))
            self.assertEqual(points[1].date, date(2022, 5, 2))
            self.assertEqual(points[30].date, date(2022, 5, 31))

            points = Point.search([
                    ('date', '>=', date(2022, 5, 1)),
                    ('date', '<=', date(2022, 5, 31)),
                ], order=[('date', 'DESC')])
            self.assertEqual(len(points), 31)
            self.assertEqual(points[0].date, date(2022, 5, 31))
            self.assertEqual(points[1].date, date(2022, 5, 30))
            self.assertEqual(points[30].date, date(2022, 5, 1))

    @with_transaction()
    def test_reporting_scaling_value(self):
        """ create chart, add graphdef, test scaling-values
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')
        GraphDef = pool.get('diagram.graphdef')
        Currency = pool.get('currency.currency')

        company = self.prep_company()
        with Transaction().set_context({
                'company': company.id}):
            chart, = Chart.create([{
                'name': 'Chart 1',
                'numdays': '30',
                'start_date': date(2022, 4, 1),
                'scaling': 'view',
                'graphdef': [('create', [{
                        'name': 'Euro',
                        'dtype': 'currency.currency',
                        'currency': company.currency.id,
                    }, ])],
                }])

            # set date of existing rate to 2022-04-01
            Currency.write(*[
                [company.currency],
                {
                    'rates': [
                        ('write',
                            [company.currency.rates[0]], {
                                'date': date(2022, 4, 1),
                            })],
                }])

            self.assertEqual(len(company.currency.rates), 1)
            self.assertEqual(company.currency.rates[0].rate, Decimal('1.0'))
            self.assertEqual(company.currency.rates[0].date, date(2022, 4, 1))

            # add 2x rates
            Currency.write(*[
                [company.currency],
                {
                    'rates': [('create', [{
                        'rate': Decimal('2.0'),
                        'date': date(2022, 4, 30),
                        }, {
                        'rate': Decimal('5.0'),
                        'date': date(2022, 3, 1),
                        }])],
                }])
            self.assertEqual(len(company.currency.rates), 3)
            self.assertEqual(company.currency.rates[0].rate, Decimal('2.0'))
            self.assertEqual(company.currency.rates[0].date, date(2022, 4, 30))
            self.assertEqual(company.currency.rates[1].rate, Decimal('1.0'))
            self.assertEqual(company.currency.rates[1].date, date(2022, 4, 1))
            self.assertEqual(company.currency.rates[2].rate, Decimal('5.0'))
            self.assertEqual(company.currency.rates[2].date, date(2022, 3, 1))

            self.assertEqual(company.currency.name, 'usd')
            self.assertEqual(chart.numdays, '30')
            self.assertEqual(chart.start_date, date(2022, 4, 1))

            Chart.write(*[
                [chart],
                {
                    'scaling': 'off',
                }])
            self.assertEqual(chart.graphdef[0].scaling, 'off')
            self.assertEqual(chart.graphdef[0].get_scaling_params(), None)
            self.assertEqual(chart.lines[0].date, date(2022, 4, 1))
            # scaling off: check values
            self.assertEqual(chart.lines[0].date, date(2022, 4, 1))
            self.assertEqual(chart.lines[0].graph1, Decimal('1.0'))
            self.assertEqual(chart.lines[19].date, date(2022, 4, 20))
            self.assertEqual(chart.lines[19].graph1, Decimal('1.6552'))
            self.assertEqual(chart.lines[29].date, date(2022, 4, 30))
            self.assertEqual(chart.lines[29].graph1, Decimal('2.0'))

            # scale visible range: 1. apr - 30. apr
            Chart.write(*[
                [chart],
                {
                    'scaling': 'view',
                }])
            self.assertEqual(chart.graphdef[0].scaling, 'view')
            self.assertEqual(chart.used_start_date(), date(2022, 4, 1))
            self.assertEqual(chart.used_end_date(), date(2022, 5, 1))
            self.assertEqual(chart.graphdef[0].get_scaling_params(), {
                'factor': Decimal('100.0'),
                'offset': Decimal('-100.0'),
                'min': Decimal('0.0'),
                'max': Decimal('100.0'),
                })

            # scale visible range + 6 month before: 1 .nov - 30. apr
            Chart.write(*[
                [chart],
                {
                    'scaling': 'six',
                }])
            self.assertEqual(chart.graphdef[0].scaling, 'six')
            self.assertEqual(chart.graphdef[0].get_scaling_params(), {
                'factor': Decimal('25.0'),
                'offset': Decimal('-25.0'),
                'min': Decimal('0.0'),
                'max': Decimal('100.0'),
                })

            # delete rates, get_caling_param() results to None
            Currency.write(*[
                [company.currency],
                {
                    'rates': [('delete', company.currency.rates)],
                }])

            # no rates to compute scaling,
            # but its still in cache, get cached data
            self.assertEqual(chart.graphdef[0].get_scaling_params(), {
                'factor': Decimal('25.0'),
                'offset': Decimal('-25.0'),
                'min': Decimal('0.0'),
                'max': Decimal('100.0'),
                })

            # no rates, no cache, get None
            GraphDef._scaling_graph_cache.clear_all()
            self.assertEqual(chart.graphdef[0].get_scaling_params(), None)

    @with_transaction()
    def test_reporting_scaling_setting(self):
        """ create chart, add graphdef, test scaling-setting
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')
        GraphDef = pool.get('diagram.graphdef')

        company = self.prep_company()
        with Transaction().set_context({
                'company': company.id}):
            chart, = Chart.create([{
                'name': 'Chart 1',
                'numdays': '90',
                'start_date': date(2022, 4, 1),
                'scaling': 'view',
                'graphdef': [('create', [{
                        'name': 'Euro',
                        'dtype': 'currency.currency',
                        'currency': company.currency.id,
                        'scaling': 'alldata'
                    }, {
                        'name': 'US$',
                        'dtype': 'currency.currency',
                        'currency': company.currency.id,
                    }])],
                }])

            self.assertEqual(chart.name, 'Chart 1')
            self.assertEqual(chart.numdays, '90')
            self.assertEqual(chart.start_date, date(2022, 4, 1))
            self.assertEqual(chart.scaling, 'view')
            self.assertEqual(len(chart.graphdef), 2)
            self.assertEqual(chart.graphdef[0].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[0].scaling, 'view')
            self.assertEqual(chart.graphdef[1].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[1].scaling, 'view')

            # update 'scaling' of chart, should update graphs
            Chart.write(*[
                [chart],
                {
                    'scaling': 'six',
                }])
            self.assertEqual(chart.scaling, 'six')
            self.assertEqual(chart.graphdef[0].scaling, 'six')
            self.assertEqual(chart.graphdef[1].scaling, 'six')

            # update 'scaling' of graph, is not stored
            # because parent-scaling is not 'graph'
            GraphDef.write(*[
                [chart.graphdef[0]],
                {
                    'scaling': 'alldata',
                }])
            self.assertEqual(chart.scaling, 'six')
            self.assertEqual(chart.graphdef[0].scaling, 'six')
            self.assertEqual(chart.graphdef[1].scaling, 'six')

            # set to individial setting, per 'graph'
            Chart.write(*[
                [chart],
                {
                    'scaling': 'graph',
                }])
            self.assertEqual(chart.scaling, 'graph')
            self.assertEqual(chart.graphdef[0].scaling, 'six')
            self.assertEqual(chart.graphdef[1].scaling, 'six')

            GraphDef.write(*[
                [chart.graphdef[0]],
                {
                    'scaling': 'alldata',
                }])
            self.assertEqual(chart.scaling, 'graph')
            self.assertEqual(chart.graphdef[0].scaling, 'alldata')
            self.assertEqual(chart.graphdef[1].scaling, 'six')

    @with_transaction()
    def test_reporting_startdate_by_context(self):
        """ create chart, add graphdef, set interval,
            test selection of start-date
        """
        pool = Pool()
        Chart = pool.get('diagram.chart')
        Point = pool.get('diagram.point')

        company = self.prep_company()
        (usd, euro) = self.prep_2nd_currency(company)
        self.assertEqual(len(usd.rates), 1)
        self.assertEqual(usd.rates[0].date, date(2022, 5, 2))
        self.assertEqual(usd.rates[0].rate, Decimal('1.05'))

        with Transaction().set_context({
                'company': company.id}):
            chart, = Chart.create([{
                'name': 'Chart 1',
                'numdays': '90',
                'start_date': date(2022, 4, 1),
                'graphdef': [('create', [{
                        'name': 'Euro',
                        'dtype': 'currency.currency',
                        'currency': euro.id,
                    }, {
                        'name': 'US$',
                        'dtype': 'currency.currency',
                        'currency': usd.id,
                    }])],
                }])

            self.assertEqual(chart.name, 'Chart 1')
            self.assertEqual(chart.numdays, '90')
            self.assertEqual(chart.start_date, date(2022, 4, 1))

            self.assertEqual(chart.used_start_date(), date(2022, 4, 1))
            self.assertEqual(chart.used_end_date(), date(2022, 6, 30))

            self.assertEqual(len(chart.graphdef), 2)
            self.assertEqual(chart.graphdef[0].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[0].currency.rec_name, 'Euro')
            self.assertEqual(chart.graphdef[1].dtype, 'currency.currency')
            self.assertEqual(chart.graphdef[1].currency.rec_name, 'usd')

            # point-date selected by chart
            self.assertEqual(len(chart.lines), 90)
            self.assertEqual(chart.lines[0].date, date(2022, 4, 1))

        with Transaction().set_context({
                'company': company.id,
                'point_date': date(2022, 6, 15)}):
            # start_date is now defined by context
            chart, = Chart.search([])
            self.assertEqual(len(chart.lines), 90)
            self.assertEqual(chart.used_start_date(), date(2022, 6, 15))
            self.assertEqual(chart.used_end_date(), date(2022, 9, 13))
            self.assertEqual(chart.lines[0].date, date(2022, 6, 15))
            self.assertEqual(chart.lines[0].sequence, 0)
            self.assertEqual(chart.lines[30].date, date(2022, 7, 15))
            self.assertEqual(chart.lines[30].sequence, 30)

            points = Point.search([('date', '=', date(2022, 6, 30))])
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0].date, date(2022, 6, 30))
            self.assertEqual(points[0].sequence, 15)

        # remove start-date from chart
        chart, = Chart.search([])
        Chart.write(*[
            [chart],
            {
                'start_date': None,
            }])

        with Transaction().set_context({
                'company': company.id}):
            # start_date is now 'current-date'
            chart, = Chart.search([])
            self.assertEqual(len(chart.lines), 90)
            self.assertEqual(chart.start_date, None)

            self.assertEqual(
                chart.used_start_date(),
                date.today() - timedelta(days=90))
            self.assertEqual(chart.used_end_date(), date.today())

            start_date = date.today() - timedelta(days=int(chart.numdays))

            self.assertEqual(chart.lines[0].date, start_date)
            self.assertEqual(chart.lines[0].sequence, 0)
            self.assertEqual(
                chart.lines[30].date,
                start_date + timedelta(days=30))
            self.assertEqual(chart.lines[30].sequence, 30)

            points = Point.search([
                ('date', '=', start_date + timedelta(days=20))])
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0].date, start_date + timedelta(days=20))
            self.assertEqual(points[0].sequence, 20)

# end ChartTestCase


del ModuleTestCase
